import argparse
from connection import Server
import msgs
import os
from select import select
import subprocess
import sys
import time


next_unique_id = 0

def get_unique_id():
  global next_unique_id
  unique_id = next_unique_id
  next_unique_id += 1
  return unique_id


class Process(object):

  def __init__(self, conn, command, bufsize, service):
    super(Process, self).__init__()
    self.session = get_unique_id()
    self.conn = conn
    self.popen = subprocess.Popen(
        command,
        shell = True,
        stdin = subprocess.PIPE,
        stdout = subprocess.PIPE,
        stderr = subprocess.STDOUT,
        env = self.make_env(service)
        )
    self.bufsize = bufsize
    self.conn_closed = False
    self.input_closed = False
    self.output_closed = False
    self.retcode_sent = False

  def make_env(self, service):
    proc_env = dict(os.environ)
    proc_env["RESHD_SESSION"] = str(self.session)
    proc_env["RESHD_SERVICE"] = repr(service)
    return proc_env

  def active_files(self):
    af = []
    if not self.conn_closed and not self.input_closed:
      af.append(self.conn)
    if not self.popen.stdout.closed:
      af.append(self.popen.stdout)
    return af

  def process_io(self, ready_set, upload_table, download_table):
    if self.conn in ready_set:
      self.process_input(upload_table, download_table)
    if self.popen.stdout in ready_set:
      self.process_output()

  def process_input(self, upload_table, download_table):
    try:
      msg = msgs.recv(self.conn)
      if msg["code"] == "data":
        self.popen.stdin.write(msg["data"])
      elif msg["code"] == "close":
        self.popen.stdin.close()
      elif msg["code"] == "xfer":
        if msg["xfer"] in upload_table:
          upload_table[msg["xfer"]].forward(msg)
        elif msg["xfer"] in download_table:
          download_table[msg["xfer"]].echo(msg)
    except msgs.Error:
      self.conn.close()
      self.conn_closed = True

  def process_output(self):
    out_data = os.read(self.popen.stdout.fileno(), self.bufsize)
    try:
      if len(out_data) < 1:
        msgs.send(self.conn, code = "close")
        self.popen.stdout.close()
      else:
        msgs.send(self.conn, code = "data", data = out_data)
    except msgs.Error:
      self.conn.close()
      self.conn_closed = True

  def send_retcode_once(self):
    if not self.retcode_sent:
      retcode = self.popen.poll()
      if retcode is not None:
        msgs.send(self.conn, code = "terminated", retcode = retcode)
        self.retcode_sent = True

  def kill(self, term_timeout):
    if not self.retcode_sent:
      try:
        self.popen.terminate()
        timeout = time.time() + term_timeout
        while self.popen.poll() is None:
          time.sleep(0.1)
          if time.time() > timeout:
            self.popen.kill()
            break
        self.popen.wait()
      except OSError:
        pass


class Upload(object):

  def __init__(self, local, proc, paths):
    super(Upload, self).__init__()
    self.xfer = get_unique_id()
    self.local = local
    self.closed = False
    msgs.send(
        proc.conn, 
        code  = "upload",
        xfer  = self.xfer,
        paths = paths
        )

  def forward(self, msg):
    try:
      msgs.send(self.local, **msg)
    except msgs.Error:
      self.local.close()
      self.closed = True

  def is_closed(self, ready_fd):
    if self.closed:
      return True
    if self.local in ready_fd:
      self.local.close()
      self.closed = True
      return True
    return False


class Download(object):

  def __init__(self, local, proc, path):
    super(Download, self).__init__()
    self.xfer = get_unique_id()
    self.local = local
    self.closed = False
    self.remote = proc.conn
    msgs.send(
        proc.conn,
        code = "download",
        xfer = self.xfer,
        path = path
        )

  def active_local(self):
    return [] if self.is_closed() else [self.local]

  def is_closed(self): return self.closed

  def echo(self, msg):
    msgs.send(self.local, **msg)

  def forward(self, ready):
    if self.local in ready:
      msg = msgs.recv(self.local)
      msgs.send(self.remote, **msg)
      if "data" not in msg:
        self.local.close()
        self.closed = True


def set_up_listener(service):
  return Server(service)

def start_process(cmd, conn, proc_table, default_cmd, service):
  proc_table.append(
      Process(
        conn, 
        cmd["command"] if cmd["command"] else default_cmd,
        cmd["bufsize"],
        service
        )
      )

def get_proc_for_session(session_id, proc_table, conn):
  proc_for_session = [p for p in proc_table if p.session == session_id]
  if len(proc_for_session) == 0:
    msgs.send(conn, error = "session")
    conn.close()
    return None
  return proc_for_session[0]

def start_upload(msg, conn, upload_table, proc_table):
  proc = get_proc_for_session(msg["session"], proc_table, conn)
  upload = Upload(conn, proc, msg["paths"])
  upload_table[upload.xfer] = upload

def start_download(msg, conn, download_table, proc_table):
  proc = get_proc_for_session(msg["session"], proc_table, conn)
  download = Download(conn, proc, msg["path"])
  download_table[download.xfer] = download

def serve_forever(service, default_cmd):
  sock = set_up_listener(service)
  proc_table = []
  upload_table = {}
  download_table = {}
  while True:
    ready, _, _ = select(
        [sock] + sum([p.active_files() for p in proc_table], []) + sum(
          [d.active_local() for d in download_table.itervalues()],
          []
          ),
        [],
        [],
        1.0
        )
    if sock in ready:
      conn, _ = sock.accept()
      try:
        msg = msgs.recv(conn)
        if msg["code"] == "command":
          start_process(msg, conn, proc_table, default_cmd, service)
        elif msg["code"] == "lcat":
          start_upload(msg, conn, upload_table, proc_table)
        elif msg["code"] == "ltee":
          start_download(msg, conn, download_table, proc_table)
      except msgs.Error:
        conn.close()
    for proc in proc_table:
      proc.process_io(ready, upload_table, download_table)
      proc.send_retcode_once()
    for download in download_table.itervalues():
      download.forward(ready)
    for xfer in [x for x, u in upload_table.iteritems() if u.is_closed(ready)]:
      del upload_table[xfer]
    for xfer in [x for x, d in download_table.iteritems() if d.is_closed()]:
      del download_table[xfer]
    for i in [i for i, c in enumerate(proc_table) if c.conn_closed]:
      proc_table[i].kill(5.0)
      del proc_table[i]


parser = argparse.ArgumentParser(
    description = "Simple remote shell server."
    )
parser.add_argument(
    "address",
    help = "Interface to bind the server to."
    )
parser.add_argument(
    "port",
    type = int,
    help = "Port to bind the server to."
    )
parser.add_argument(
    "-d",
    dest = "cmd",
    default = "/bin/bash",
    help = "Default command to run."
    )
ns = parser.parse_args(sys.argv[1:])
serve_forever((ns.address, ns.port), ns.cmd)


# EOF
