import os
from Queue import Queue
import random
import re
import select
import socket
import struct
import threading
import time
import zlib


FRAME_TYPE_DATA = "data"
FRAME_TYPE_ACK = "ack-"
FRAME_TYPE_CLOSE = "clos"
FRAME_TYPE_AUTH = "auth"
FRAME_TYPE_ACK_CLOSE = "ackc"

AUTH_DISCARD_LIMIT = 0
AUTH_ABANDON_LIMIT = 10
AUTH_NULL = 11
AUTH_IDENTIFIED = 12
AUTH_ANY = 1
AUTH_NEW = 2
AUTH_TRANSFER = -1
AUTH_INVALID = -2
AUTH_COMPLETE = 100

MAX_RAW_PAYLOAD_LEN = 3057
MAX_FRAME_LEN = 20 + MAX_RAW_PAYLOAD_LEN * 4 / 3

SW_TIMEOUT = 1.0


class SocketError(Exception):
  pass

class DiscardConnection(Exception):
  pass


class Connection(object):

  def __init__(self):
    self.service, self.peer = socket.socketpair()
    self.thr = threading.Thread(target = self.run)
    self.thr.daemon = True
    self.thr.start()

  def fileno(self):
    return self.service.fileno()

  def close(self):
    self.service.close()
    self.thr.join()
    self.close_socket()


def escape_payload(s):
  esc_p = []
  n = 0
  while n < len(s):
    if s[n:].startswith(">>>"):
      esc_p.append(">>><")
      n += 3
    else:
      esc_p.append(s[n])
      n += 1
  payload_escaped_almost = "".join(esc_p)
  if payload_escaped_almost.endswith(">"):
    return payload_escaped_almost + "<"
  return payload_escaped_almost

def unescape_payload(ep):
  unesc_p = []
  n = 0
  while n < len(ep):
    m = re.match(">+<", ep[n:])
    if m is None:
      unesc_p.append(ep[n])
      n += 1
    else:
      unesc_p.append(m.group(0)[:-1])
      n += len(m.group(0))
  return "".join(unesc_p)

def send_exact(sock, s):
  try:
    while len(s) > 0:
      n = sock.send(s)
      if n == 0:
        raise SocketError()
      s = s[n:]
  except socket.error:
    raise SocketError()

def wrap_uint_crc32(n):
  n_packed = struct.pack("!I", n)
  return n_packed + struct.pack("!i", zlib.crc32(n_packed))

def send_uint_crc32(sock, n):
  send_exact(sock, wrap_uint_crc32(n))

def recv_uint_crc32(sock, timeout = None):
  ready, _, _ = select.select([sock], [], [], timeout)
  if sock not in ready:
    return None
  packet = sock.recv(8)
  if len(packet) < 8:
    return None
  n, crc = struct.unpack("!Ii", packet)
  if crc != zlib.crc32(packet[0:4]):
    return None
  return n

def serialise_frame(ftype, seqno, payload_escaped):
  frame_base = (
      "<<<<" +
      ftype +
      struct.pack("!I", seqno) +
      payload_escaped +
      ">>>>"
      )
  crc = zlib.crc32(frame_base)
  return frame_base + struct.pack("!i", crc)

def send_ack(sock, seqno):
  send_exact(sock, serialise_frame(FRAME_TYPE_ACK, seqno, ""))

def send_ack_close(sock, seqno):
  send_exact(sock, serialise_frame(FRAME_TYPE_ACK_CLOSE, seqno, ""))

def send_auth(sock, conn_id):
  send_exact(sock, serialise_frame(FRAME_TYPE_AUTH, conn_id, ""))


class FrameToSend(object):

  def __init__(self, ftype, seqno, payload_raw):
    self.ftype = ftype
    self.seqno = seqno
    self.payload_escaped = escape_payload(payload_raw)
    self.timer = None

  def serialise(self):
    return serialise_frame(self.ftype, self.seqno, self.payload_escaped)

  def send(self, sock, timer_pipe, auth_status):
    self.timer = threading.Timer(SW_TIMEOUT, self.ask_resend, args = (timer_pipe,))
    self.timer.start()
    if auth_status == AUTH_COMPLETE:
      frame = self.serialise()
      send_exact(sock, self.serialise())
      return len(frame)
    return 0

  def ask_resend(self, timer_pipe):
    try:
      os.write(timer_pipe, struct.pack("!cI", "r", self.seqno))
    except OSError:
      #print "****", self.seqno, "****"
      pass
    except AttributeError:
      pass
    self.timer = None

  def acked(self):
    if self.timer is not None:
      self.timer.cancel()
      self.timer = None


class FrameData(FrameToSend):

  def __init__(self, sock_src, seqno):
    try:
      payload_raw = os.read(sock_src.fileno(), MAX_RAW_PAYLOAD_LEN)
      if len(payload_raw) == 0:
        raise SocketError()
    except OSError:
      raise SocketError()
    super(FrameData, self).__init__(FRAME_TYPE_DATA, seqno, payload_raw)


class FrameClose(FrameToSend):

  def __init__(self, seqno):
    super(FrameClose, self).__init__(FRAME_TYPE_CLOSE, seqno, "")


def parse_frame(buf):
  frame_start = buf.find("<<<<")
  if frame_start < 0:
    return None, buf[-4:]
  buf = buf[frame_start:]
  frame_start = 0
  frame_end = buf.find(">>>>", 12)
  if frame_end < 0:
    return None, buf
  frame_end += 8
  if frame_end > len(buf):
    return None, buf
  if frame_end - frame_start < 20:
    return "error", buf[frame_end:]
  frame_base = buf[frame_start : frame_end - 4]
  crc = struct.unpack("!i", buf[frame_end - 4 : frame_end])[0]
  if zlib.crc32(frame_base) != crc:
    return "error", buf[frame_end:]  # Drop error-laden frame.
  ftype = buf[frame_start + 4 : frame_start + 8]
  seqno = struct.unpack("!I", buf[frame_start + 8 : frame_start + 12])[0]
  payload_escaped = buf[frame_start + 12 : frame_end - 8]
  return FrameReceived(ftype, seqno, payload_escaped), buf[frame_end:]


class FrameReceived(object):

  def __init__(self, ftype, seqno, payload_escaped):
    super(FrameReceived, self).__init__()
    self.ftype = ftype
    self.seqno = seqno
    self.payload = unescape_payload(payload_escaped)


class Peer(Connection):

  def __init__(self):
    self.conn_id = None
    self.master_seqno = 1
    super(Peer, self).__init__()

  def close_socket(self):
    if self.socket is not None:
      self.socket.close()
      self.socket = None

  def before_connection_loop(self, timer_pipe):
    pass

  def after_connection_loop(self):
    pass

  def run(self):
    auth_status = self.init_auth()
    closing_connection = False
    connection_alive = True
    ack_close_received = False
    send_window_size = 4
    recv_window_size = 4
    send_window = []
    recv_buffer = ""
    recv_window = []
    last_frame_acked = 0
    max_seqno_acceptable = last_frame_acked + recv_window_size
    timer_recv, timer_send = os.pipe()
    closing_frame = None
    self.before_connection_loop(timer_send)

    while connection_alive:
      sock = self.get_socket()

      if closing_connection and len(send_window) == 0 and closing_frame is None:
        closing_frame = FrameClose(self.master_seqno)
        self.master_seqno += 1
        closing_frame.send(sock, timer_send, auth_status)

      try:
        ready, _, _ = select.select(
            (
              (
                [self.peer]
                if (
                  not closing_connection and
                  auth_status == AUTH_COMPLETE and
                  len(send_window) < send_window_size
                  )
                else []
                ) +
              ([sock] if sock is not None else []) +
              [timer_recv]
              ),
            [],
            []
            )
      except ValueError:
        closing_connection = True

      try:
        if self.peer in ready:  # New message to send.
          frame = None
          try:
            frame = FrameData(self.peer, self.master_seqno)
            self.master_seqno += 1
          except SocketError:
            closing_connection = True
          else:
            send_window.append(frame)
            frame.send(sock, timer_send, auth_status)

        if sock in ready:
          try:
            recv_bytes = os.read(sock.fileno(), MAX_FRAME_LEN)
            if len(recv_bytes) == 0:
              if ack_close_received:
                connection_alive = False
                break
              else:
                raise SocketError()
          except OSError:
            raise SocketError()
          recv_buffer += recv_bytes
          while True:
            frame, recv_buffer = parse_frame(recv_buffer)
            if frame is None:
              break  # Not enough bytes to form a frame.
            if frame == "error":
              continue  # Erroneous frame from buffer prefix.

            if frame.ftype == FRAME_TYPE_DATA and auth_status == AUTH_COMPLETE:
              if frame.seqno > last_frame_acked and frame.seqno <= max_seqno_acceptable:
                # Frame is within window.
                if all(f.seqno != frame.seqno for f in recv_window):
                  recv_window.append(frame)
                  recv_window.sort(key = lambda f: f.seqno)
                while (
                    len(recv_window) > 0 and
                    recv_window[0].seqno == last_frame_acked + 1
                    ):
                  last_frame_acked += 1
                  max_seqno_acceptable += 1
                  frame_acked = recv_window.pop(0)
                  try:
                    if not closing_connection:
                      send_exact(self.peer, frame_acked.payload)
                  except SocketError:
                    closing_connection = True
              send_ack(sock, last_frame_acked)

            elif frame.ftype == FRAME_TYPE_ACK and auth_status == AUTH_COMPLETE:
              if closing_frame and frame.seqno == closing_frame.seqno:
                # Other side acknowledges connection end.
                closing_frame.acked()
                connection_alive = False
              else:
                while (
                    len(send_window) > 0 and
                    send_window[0].seqno <= frame.seqno
                    ):
                  send_window[0].acked()
                  del send_window[0]

            elif frame.ftype == FRAME_TYPE_CLOSE and auth_status == AUTH_COMPLETE:
              send_ack_close(sock, frame.seqno)

            elif frame.ftype == FRAME_TYPE_ACK_CLOSE:
              if closing_frame is None:
                # Closure responder: ack-close received by initiator; he's
                # ready for us to drop.
                connection_alive = False
              else:
                # Closure initiator: send another ack-close, wait for other
                # end to drop.
                if ack_close_received:
                  # Already received ack-close, and now another -- this means
                  # that both sides are close initiators. Ack-closes have been
                  # properly exchanged, now drop.
                  connection_alive = False
                else:
                  send_ack_close(sock, frame.seqno)
                  ack_close_received = True

            elif frame.ftype == FRAME_TYPE_AUTH:
              conn_id = frame.seqno  # Seqno co-opted to pass on connection ID.
              auth_status = self.handle_auth(conn_id, sock, auth_status)
              if auth_status < AUTH_DISCARD_LIMIT:
                raise DiscardConnection()

            else:
              print "*** Dropping frame of type", frame.ftype, "***"

        if timer_recv in ready:  # Must resend a message or initiate reconnection.
          action = struct.unpack("!c", os.read(timer_recv, 1))[0]
          if action == "r":
            seqno = struct.unpack("!I", os.read(timer_recv, 4))[0]
            for frame in (f for f in send_window if f.seqno == seqno):
              frame.send(sock, timer_send, auth_status)
            if closing_frame is not None and closing_frame.seqno == seqno:
              closing_frame.send(sock, timer_send, auth_status)
          elif action == "x":
            try:
              if auth_status != AUTH_COMPLETE:
                if self.socket is None:
                  self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM, 0)
                  self.socket.connect(self.addr_port)
                send_auth(self.socket, self.conn_id)
            except Exception:
              raise
            finally:
              self.set_up_connect_timer(timer_send)

      # Lost socket! Will recover next iteration.
      except socket.error:
        self.close_socket()
        auth_status = self.reset_auth(auth_status)
        if auth_status < AUTH_ABANDON_LIMIT:
          break
      except SocketError:
        self.close_socket()
        auth_status = self.reset_auth(auth_status)
        if auth_status < AUTH_ABANDON_LIMIT:
          break
      except DiscardConnection:
        break

    # All done!
    os.close(timer_send)
    os.close(timer_recv)
    if not connection_alive:
      self.after_connection_loop()
      self.close_socket()

  def send(self, s):
    return self.service.send(s)

  def recv(self, n):
    return self.service.recv(n)


class Client(Peer):

  def __init__(self, addr_port):
    self.addr_port = addr_port
    self.socket = None
    self.connect_timer = None
    super(Client, self).__init__()

  def init_auth(self):
    return AUTH_NULL

  def reset_auth(self, auth_status):
    if auth_status == AUTH_COMPLETE:
      return AUTH_IDENTIFIED
    return auth_status

  def get_socket(self):
    return self.socket

  def before_connection_loop(self, timer_pipe):
    self.conn_id = 0
    self.ask_reconnect(timer_pipe)  # As if the timer fired right now.

  def set_up_connect_timer(self, timer_pipe):
    self.connect_timer = threading.Timer(
        SW_TIMEOUT,
        self.ask_reconnect,
        args = [timer_pipe]
        )
    self.connect_timer.start()

  def ask_reconnect(self, timer_pipe):
    os.write(timer_pipe, "x")

  def after_connection_loop(self):
    if self.connect_timer:
      self.connect_timer.cancel()

  def handle_auth(self, conn_id, sock, auth_status):
    if auth_status == AUTH_NULL or (
        auth_status == AUTH_IDENTIFIED and conn_id != self.conn_id
        ):
      self.conn_id = conn_id
      send_auth(self.socket, self.conn_id)
      return AUTH_IDENTIFIED
    return AUTH_COMPLETE


class Accepted(Peer):

  def __init__(self, sock, peer, server):
    self.socket_ready = threading.Event()
    self.set_socket(sock)
    self.peer = peer
    self.server = server
    super(Accepted, self).__init__()

  def init_auth(self):
    return AUTH_ANY

  def reset_auth(self, auth_status):
    return auth_status

  def get_socket(self):
    self.socket_ready.wait()
    return self.socket

  def close_socket(self):
    self.socket_ready.clear()
    super(Accepted, self).close_socket()

  def set_socket(self, sock, auth_status = None):
    self.socket = sock
    self.socket_ready.set()

  def handle_auth(self, conn_id, sock, auth_status):
    if auth_status == AUTH_ANY:
      if conn_id == 0 or not self.server.is_active_conn(conn_id):
        self.conn_id = self.server.gen_conn_id()
        send_auth(sock, self.conn_id)
        return AUTH_NEW
      elif self.server.is_active_conn(conn_id):
        send_auth(sock, conn_id)
        self.server.transfer_socket(conn_id, sock)
        return AUTH_TRANSFER
    elif auth_status == AUTH_NEW:
      if self.conn_id == conn_id:
        send_auth(sock, conn_id)
        self.server.add_client(self)
        return AUTH_COMPLETE
      elif conn_id == 0:
        # Client did not receive the initial auth response.
        send_auth(sock, self.conn_id)
        return AUTH_NEW
      else:
        raise RuntimeError("NEW connection received Auth with differing code.")
    elif auth_status == AUTH_COMPLETE:
      # Ack the AUTH with another one.
      if self.conn_id != conn_id:
        raise RuntimeError("NEW connection received Auth with differing code.")
      send_auth(sock, conn_id)
      return AUTH_COMPLETE

  def after_connection_loop(self):
    self.server.remove_client(self.conn_id)


class Server(Connection):

  def __init__(self, addr_port):
    self.server = socket.socket(socket.AF_INET, socket.SOCK_STREAM, 0)
    self.server.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
    self.server.bind(addr_port)
    self.server.listen(5)
    self.accept_queue = Queue()
    self.active_conn = {}
    super(Server, self).__init__()

  def run(self):
    while True:
      try:
        ready, _, _ = select.select([self.server, self.peer], [], [])
      except ValueError:
        break  # Interface was closed, so shut down.
      if self.peer in ready:
        # Only used to ask for closure.
        break
      if self.server in ready:
        try:
          sock, peer = self.server.accept()
          Accepted(sock, peer, self)
        except socket.error:
          # By closing the peer interface, we'll force an error on a
          # service interface read, which will finish the shutdown
          # in turn.
          self.peer.shutdown(socket.SHUT_RDWR)
          self.peer.close()

  def gen_conn_id(self):
    conn_id = 0
    while conn_id == 0 or self.is_active_conn(conn_id):
      conn_id = random.getrandbits(32)
    return conn_id

  def is_active_conn(self, conn_id):
    return conn_id in self.active_conn

  def add_client(self, acc):
    if self.is_active_conn(acc.conn_id):
      return False
    self.active_conn[acc.conn_id] = acc
    self.accept_queue.put(acc)
    self.peer.send("0")
    return True

  def remove_client(self, conn_id):
    if conn_id in self.active_conn:
      del self.active_conn[conn_id]

  def transfer_socket(self, conn_id, sock):
    if conn_id not in self.active_conn:
      raise RuntimeError("Connection ID does not correspond to active connection!")
    self.active_conn[conn_id].set_socket(sock, AUTH_COMPLETE)

  def accept(self):
    # Consume the byte used to wake up the service interface. If something
    # bad happened and the interface got closed in the meantime, this
    # should raise an exception anyhow.
    try:
      self.service.recv(1)
      acc = self.accept_queue.get_nowait()
      return acc, acc.peer
    except:
      raise SocketError()

  def close_socket(self):
    self.server.close()


# EOF
