import argparse
import os
import os.path
import Queue
import random
import re
import select
import shlex
import shutil
import signal
from subprocess import Popen, PIPE
import sys
import threading
import time
import traceback

FILE = 1
class TestCase(object):

    code_root = None
    unique_id = 0

    @classmethod
    def get_unique_id(cls):
        the_id = cls.unique_id
        cls.unique_id += 1
        return the_id

    @classmethod
    def code_root_join(self, *path):
        return os.path.join(TestCase.code_root, *path)

    @classmethod
    def make_server(self, addr, port):
        return Popen(
                ["./reshd", addr, str(port)],
                bufsize= 0,
                cwd = self.code_root_join("server")
                )

    @classmethod
    def make_client(self, addr, port, cmd, output):
        return Popen(
                ["./resh", addr, str(port)] + (["/bin/bash"] if cmd is None else shlex.split(cmd)),
                cwd    = self.code_root_join("client"),
                stdin  = PIPE,
                stdout = output
                )

    def addr(self):
        return "localhost"

    def port(self):
        return 9887

    def client_list(self):
        return [("/bin/bash", sys.stdout)]

    def local_files(self):
        return []

    def local_path(self, name):
        return self.code_root_join("client", name)

    def local_paths(self):
        return [self.local_path(n) for n in self.local_files()]

    def remote_files(self):
        return []

    def remote_path(self, name):
        return self.code_root_join("server", name)

    def remote_paths(self):
        return [self.remote_path(n) for n in self.remote_files()]

    def run(self):
        try:
            try:
                self.set_up()
            except:
                print "***** ERROR in setup *****"
                raise
            try:
                self.test()
            except:
                print "***** TEST FAILURE *****"
                raise
        except Exception as ex:
            print str(ex)
        finally:
            try:
                self.tear_down()
            except Exception as ex:
                print "***** ERROR in teardown *****"
                print str(ex)

    def set_up(self):
        self.server = self.make_server(self.addr(), self.port())
        time.sleep(0.2)
        self.output_data = [
                dict(
                    file = (
                        open(self.code_root_join("out-{}".format(self.get_unique_id())), "w")
                        if o == FILE else o
                        ),
                    expected = None
                    )
                for _, o in self.client_list()
                ]
        self.clients = [
                self.make_client(self.addr(), self.port(), cmd, od["file"])
                for (cmd, _), od in zip(self.client_list(), self.output_data)
                ]

    def kill_in_steps(self, p, timeout_close, timeout_term):
        if p.poll() is None and timeout_close is not None:
            p.stdin.close()
            self.poll_death(p, timeout_close)
        if p.poll() is None and timeout_term is not None:
            p.terminate()
            self.poll_death(p, timeout_term)
        if p.poll() is None:
            p.kill()
            p.wait()

    def poll_death(self, p, timeout_limit):
        timeout = time.time() + timeout_limit
        while time.time() < timeout:
            if p.poll() is not None:
                return p.returncode
            time.sleep(0.1)
        return p.poll()

    def tear_down(self):
        for i, (c, od) in enumerate(zip(self.clients, self.output_data)):
            self.kill_in_steps(c, 5.0, 5.0)
            if "out-" in od["file"].name:
                od["file"].close()
                print "Client {} (output file {}):".format(i, od["file"].name)
                if od["expected"] is not None:
                    print_long_output("Expected", od["expected"])
                with open(od["file"].name) as out_file:
                    print_long_output("Obtained", out_file.read())
                os.remove(od["file"].name)
        self.kill_in_steps(self.server, None, 5.0)
        for p in self.local_paths() + self.remote_paths():
            if os.path.exists(p):
                os.remove(p)

    def cmd(self, client, cmd, expected_output, timeout = 0.5):
        print "[Input:  {}]".format(cmd.rstrip())
        print "[Expect: {}]".format(repr(expected_output))
        client.stdin.write(cmd)
        time.sleep(timeout)


def print_long_output(title, output):
    print "*** {} -- {} bytes ***".format(title, len(output))
    if len(output) > 128:
        print repr(output[0:30]), "...", repr(output[-30:])
    else:
        print repr(output)

def prepare_client_server(code_root):
    for sd in ["client", "server"]:
        if os.path.isdir(sd):
            shutil.rmtree(sd)
        shutil.copytree(code_root, sd)

def clean_up():
    for sd in ["client", "server"]:
        shutil.rmtree(sd)

def run_tests(argv, loc_symbols):
    parser = argparse.ArgumentParser(
            description = "Test runner"
            )
    parser.add_argument(
            "code_root",
            help = "Chemin vers le repertoire contenant le code a tester, organise en sous-repertoires 'server' et 'client'."
            )
    parser.add_argument(
            "tests",
            nargs = "*",
            help = "Tests a executer."
            )
    ns = parser.parse_args(argv)
    prepare_client_server(ns.code_root)
    TestCase.code_root = "."
    os.putenv("PYTHONUNBUFFERED", "yes")
    test_list = [
            name
            for name in loc_symbols.iterkeys()
            if name.startswith("Test") and not name.startswith("TestCase")
            ]
    if len(ns.tests) > 0:
        test_list = ns.tests
    for test_name in sorted(test_list):
        print "=" * len(test_name)
        print test_name
        print "=" * len(test_name)
        pid = os.fork()
        if pid == 0:
            # Child: run the test.
            loc_symbols[test_name]().run()
            sys.exit(0)
        else:
            # Wait for the test to complete.
            test_complete = False
            timeout = time.time() + 40.0
            while time.time() < timeout:
                pid_wait, exit_code = os.waitpid(pid, os.WNOHANG)
                if pid_wait == pid:
                    test_complete = True
                    break
                time.sleep(0.2)
            if not test_complete:
                print "***** ABORT TEST *****"
                os.kill(pid, signal.SIGINT)
                os.waitpid(pid, 0)
    clean_up()


# EOF
